#!/usr/bin/env python


import importlib.util
import sys
import types
import unittest
from unittest import mock

import stone


class TestVersion(unittest.TestCase):

    def test_generated_version_is_used_in_source_checkout(self):
        generated_version = types.ModuleType('stone._version')
        generated_version.__version__ = '9.8.7'

        spec = importlib.util.spec_from_file_location(
            'stone._version_test', stone.__file__)
        module = importlib.util.module_from_spec(spec)
        module.__package__ = 'stone'
        with mock.patch.dict(sys.modules, {'stone._version': generated_version}):
            spec.loader.exec_module(module)

        self.assertEqual(module.__version__, '9.8.7')


if __name__ == '__main__':
    unittest.main()



import inspect
import unittest

from stone._vendor.ply import yacc


class TestVendoredPly(unittest.TestCase):

    def test_pickle_table_support_is_removed(self):
        self.assertNotIn('picklefile', inspect.signature(yacc.yacc).parameters)
        self.assertFalse(hasattr(yacc.LRTable, 'read_pickle'))
        self.assertFalse(hasattr(yacc.LRGeneratedTable, 'pickle_table'))

    def test_picklefile_argument_is_rejected(self):
        with self.assertRaises(TypeError):
            yacc.yacc(picklefile='untrusted.pkl')  # pylint: disable=unexpected-keyword-arg


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python


import contextlib
import io
import json
import os
import tempfile
import unittest

from stone import __version__
from stone.cli import (
    _actual_outputs,
    _cmdline_parser,
    _load_expected_output_manifest,
    _recursive_stone_specs,
    _validate_expected_output_manifest,
)
from stone.cli_helpers import parse_route_attr_filter


class MockRoute():
    """Used to test filtering on a route's attrs."""

    def __init__(self, attrs):
        self.attrs = attrs


class TestCLI(unittest.TestCase):

    def test_parse_route_attr_filter(self):

        _, errs = parse_route_attr_filter('*=3')
        self.assertNotEqual(len(errs), 0)

        _, errs = parse_route_attr_filter('test')
        self.assertEqual(len(errs), 1)
        self.assertEqual(errs[0], 'Unexpected end of expression.')

        _, errs = parse_route_attr_filter('hide=true)')
        self.assertNotEqual(len(errs), 0)

        _, errs = parse_route_attr_filter('(hide=true')
        self.assertNotEqual(len(errs), 0)

        _, errs = parse_route_attr_filter('hide=true and and size=1')
        self.assertNotEqual(len(errs), 0)

        # Test bool
        expr, errs = parse_route_attr_filter('hide=true')
        self.assertEqual(len(errs), 0)
        self.assertTrue(expr.eval(MockRoute({'hide': True})))
        self.assertFalse(expr.eval(MockRoute({'hide': 'true'})))

        # Test int
        expr, errs = parse_route_attr_filter('level=1')
        self.assertEqual(len(errs), 0)
        self.assertTrue(expr.eval(MockRoute({'level': 1})))
        self.assertFalse(expr.eval(MockRoute({'level': 2})))
        self.assertFalse(expr.eval(MockRoute({'level': '1'})))
        self.assertFalse(expr.eval(MockRoute({})))

        # Test float
        expr, errs = parse_route_attr_filter('f=1.25')
        self.assertEqual(len(errs), 0)
        self.assertTrue(expr.eval(MockRoute({'f': 1.25})))
        self.assertFalse(expr.eval(MockRoute({'f': 3})))
        self.assertFalse(expr.eval(MockRoute({'f': '1.25'})))
        self.assertFalse(expr.eval(MockRoute({})))

        # Test string
        expr, errs = parse_route_attr_filter('status="alpha"')
        self.assertEqual(len(errs), 0)
        self.assertTrue(expr.eval(MockRoute({'status': 'alpha'})))
        self.assertFalse(expr.eval(MockRoute({'status': 'beta'})))
        self.assertFalse(expr.eval(MockRoute({'status': 0})))
        self.assertFalse(expr.eval(MockRoute({})))

        # Test null
        expr, errs = parse_route_attr_filter('status=null')
        self.assertEqual(len(errs), 0)
        self.assertTrue(expr.eval(MockRoute({'status': None})))
        self.assertFalse(expr.eval(MockRoute({'status': 'beta'})))
        self.assertFalse(expr.eval(MockRoute({'status': 0})))
        self.assertTrue(expr.eval(MockRoute({})))

        # Test conjunction: or
        expr, errs = parse_route_attr_filter('a=1 or b=1')
        self.assertEqual(len(errs), 0)
        self.assertTrue(expr.eval(MockRoute({'a': 1})))
        self.assertTrue(expr.eval(MockRoute({'b': 1})))
        self.assertTrue(expr.eval(MockRoute({'a': 1, 'b': 1})))
        self.assertTrue(expr.eval(MockRoute({'a': 1, 'b': 10})))
        self.assertFalse(expr.eval(MockRoute({'a': '0', 'b': 0})))
        self.assertFalse(expr.eval(MockRoute({'a': 0})))
        self.assertFalse(expr.eval(MockRoute({})))

        # Test conjunction: and
        expr, errs = parse_route_attr_filter('a=1 and b=1')
        self.assertEqual(len(errs), 0)
        self.assertTrue(expr.eval(MockRoute({'a': 1, 'b': 1})))
        self.assertFalse(expr.eval(MockRoute({'a': 1})))
        self.assertFalse(expr.eval(MockRoute({'b': 1})))
        self.assertFalse(expr.eval(MockRoute({'a': 1, 'b': 10})))
        self.assertFalse(expr.eval(MockRoute({'a': '0', 'b': 0})))
        self.assertFalse(expr.eval(MockRoute({'a': 0})))
        self.assertFalse(expr.eval(MockRoute({})))

        # Test multiple conjunctions
        expr, errs = parse_route_attr_filter('a=1 or a=2 or a=3')
        self.assertEqual(len(errs), 0)
        self.assertTrue(expr.eval(MockRoute({'a': 1})))
        self.assertTrue(expr.eval(MockRoute({'a': 2})))
        self.assertTrue(expr.eval(MockRoute({'a': 3})))
        self.assertFalse(expr.eval(MockRoute({'a': 4})))

        # Test "and" has higher precendence than "or"
        expr, errs = parse_route_attr_filter('a=1 or a=2 and b=3 and c=4')
        self.assertEqual(len(errs), 0)
        self.assertTrue(expr.eval(MockRoute({'a': 1})))
        self.assertFalse(expr.eval(MockRoute({'a': 2})))
        self.assertTrue(expr.eval(MockRoute({'a': 2, 'b': 3, 'c': 4})))
        self.assertTrue(expr.eval(MockRoute({'a': 1, 'b': 3, 'c': 4})))
        self.assertFalse(expr.eval(MockRoute({'a': 0, 'b': 3, 'c': 4})))

        expr, errs = parse_route_attr_filter('a=2 and b=3 and c=4 or a=1')
        self.assertEqual(len(errs), 0)
        self.assertTrue(expr.eval(MockRoute({'a': 1})))
        self.assertFalse(expr.eval(MockRoute({'a': 2})))
        self.assertTrue(expr.eval(MockRoute({'a': 2, 'b': 3, 'c': 4})))
        self.assertTrue(expr.eval(MockRoute({'a': 1, 'b': 3, 'c': 4})))
        self.assertFalse(expr.eval(MockRoute({'a': 0, 'b': 3, 'c': 4})))

        # Test parentheses for overriding precedence
        expr, errs = parse_route_attr_filter('(a=1 or a=2) and b=3 and c=4')
        self.assertEqual(len(errs), 0)
        self.assertTrue(expr.eval(MockRoute({'a': 1, 'b': 3, 'c': 4})))
        self.assertTrue(expr.eval(MockRoute({'a': 2, 'b': 3, 'c': 4})))
        self.assertFalse(expr.eval(MockRoute({'a': 1})))
        self.assertFalse(expr.eval(MockRoute({'a': 1, 'b': 3})))

    def test_actual_outputs(self):
        with tempfile.TemporaryDirectory() as output_root:
            os.makedirs(os.path.join(output_root, 'nested'))
            with open(os.path.join(output_root, 'Generated.py'), 'w', encoding='utf-8'):
                pass
            with open(
                    os.path.join(output_root, 'nested', 'Generated.swift'),
                    'w',
                    encoding='utf-8'):
                pass

            self.assertEqual(
                _actual_outputs(output_root),
                ['Generated.py', 'nested/Generated.swift'])

    def test_recursive_stone_specs(self):
        with tempfile.TemporaryDirectory() as spec_root:
            os.makedirs(os.path.join(spec_root, 'nested'))
            with open(os.path.join(spec_root, 'top.stone'), 'w', encoding='utf-8'):
                pass
            with open(os.path.join(spec_root, 'nested', 'child.stone'), 'w', encoding='utf-8'):
                pass
            with open(os.path.join(spec_root, 'nested', 'ignored.txt'), 'w', encoding='utf-8'):
                pass

            self.assertEqual(
                _recursive_stone_specs(spec_root),
                [
                    os.path.join(spec_root, 'nested', 'child.stone'),
                    os.path.join(spec_root, 'top.stone'),
                ])

    def test_load_expected_output_manifest(self):
        with tempfile.NamedTemporaryFile(mode='w', encoding='utf-8') as manifest_file:
            json.dump(['b.py', 'a.py'], manifest_file)
            manifest_file.flush()

            self.assertEqual(_load_expected_output_manifest(manifest_file.name), ['a.py', 'b.py'])

    def test_validate_expected_output_manifest(self):
        _validate_expected_output_manifest(['a.py'], ['a.py'])

        with self.assertRaises(SystemExit):
            _validate_expected_output_manifest(['a.py'], ['b.py'])

    def test_version_flag(self):
        # --version prints the version and exits 0, without requiring the
        # otherwise-mandatory positional arguments.
        out = io.StringIO()
        with self.assertRaises(SystemExit) as cm, contextlib.redirect_stdout(out):
            _cmdline_parser.parse_args(['--version'])
        self.assertEqual(cm.exception.code, 0)
        self.assertIn(__version__, out.getvalue())


if __name__ == '__main__':
    unittest.main()



import os
import tempfile
import types
import unittest

from stone.backend import Backend, CodeBackend
from stone.backends.swift import SwiftBaseBackend
from stone.compiler import Compiler


class TestOutputManifest(unittest.TestCase):

    def _build_manifest(self, backend_cls):
        module = types.ModuleType('manifest_backend')
        setattr(module, backend_cls.__name__, backend_cls)

        with tempfile.TemporaryDirectory() as output_root:
            compiler = Compiler(
                None,
                module,
                [],
                output_root,
                output_manifest=True)
            compiler.build()
            existing_files = []
            for root, _, file_names in os.walk(output_root):
                for file_name in file_names:
                    existing_files.append(
                        os.path.relpath(os.path.join(root, file_name), output_root))
            return compiler.output_manifest(), sorted(existing_files)

    def test_output_to_relative_path_records_without_writing(self):

        class ManifestBackend(CodeBackend):
            preserve_aliases = True

            def generate(self, api):
                with self.output_to_relative_path('Generated.py'):
                    self.emit('generated = True')

        manifest, existing_files = self._build_manifest(ManifestBackend)

        self.assertEqual(manifest, ['Generated.py'])
        self.assertEqual(existing_files, [])

    def test_copy_to_path_records_destination_without_copying(self):
        with tempfile.NamedTemporaryFile() as source_file:

            class CopyBackend(Backend):
                preserve_aliases = True
                src_path = source_file.name

                def generate(self, api):
                    resources_path = os.path.join(self.target_folder_path, 'Resources')
                    os.makedirs(resources_path)
                    self.copy_to_path(self.src_path, resources_path)

            manifest, existing_files = self._build_manifest(CopyBackend)

        self.assertEqual(manifest, ['Resources/{}'.format(os.path.basename(source_file.name))])
        self.assertEqual(existing_files, [])

    def test_swift_output_records_without_writing(self):

        class SwiftBackend(SwiftBaseBackend):
            preserve_aliases = True

            def generate(self, api):
                self._write_output_in_target_folder('final class Generated {}', 'Generated.swift')

        manifest, existing_files = self._build_manifest(SwiftBackend)

        self.assertEqual(manifest, ['Generated.swift'])
        self.assertEqual(existing_files, [])


class TestOutputRootValidation(unittest.TestCase):

    def test_output_to_relative_path_rejects_parent_paths(self):
        class ValidatingBackend(CodeBackend):
            preserve_aliases = True

            def generate(self, api):
                pass

        with tempfile.TemporaryDirectory() as output_root:
            backend = ValidatingBackend(output_root, [])

            with self.assertRaises(AssertionError):
                with backend.output_to_relative_path('../Generated.py'):
                    backend.emit('generated = True')

    def test_copy_to_path_rejects_parent_paths(self):
        class ValidatingBackend(Backend):
            preserve_aliases = True

            def generate(self, api):
                pass

        with tempfile.NamedTemporaryFile() as source_file:
            with tempfile.TemporaryDirectory() as output_root:
                backend = ValidatingBackend(output_root, [])

                with self.assertRaises(AssertionError):
                    backend.copy_to_path(
                        source_file.name,
                        os.path.join(output_root, '..', 'Copied.py'))

    def test_swift_output_rejects_parent_paths(self):
        class ValidatingBackend(SwiftBaseBackend):
            preserve_aliases = True

            def generate(self, api):
                pass

        with tempfile.TemporaryDirectory() as output_root:
            backend = ValidatingBackend(output_root, [])

            with self.assertRaises(AssertionError):
                backend._write_output_in_target_folder(
                    'final class Generated {}',
                    '../Generated.swift')


if __name__ == '__main__':
    unittest.main()



import os
import shutil
import tempfile
import textwrap

from stone.backends.python_client import PythonClientBackend
from stone.ir import (
    Api,
    ApiNamespace,
    ApiRoute,
    Int32,
    List,
    Map,
    Nullable,
    String,
    Void,
    StructField,
    Struct,
)

MYPY = False
if MYPY:
    import typing  # noqa: F401 # pylint: disable=import-error,unused-import,useless-suppression

import unittest

class TestGeneratedPythonClient(unittest.TestCase):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def _evaluate_namespace(self, ns):
        # type: (ApiNamespace) -> typing.Text

        backend = PythonClientBackend(
            target_folder_path='output',
            args=['-a', 'scope', '-a', 'another_attribute',
                  '-m', 'files', '-c', 'DropboxBase', '-t', 'dropbox'])
        backend._generate_routes(ns)
        return backend.output_buffer_to_string()

    def _evaluate_namespace_with_auth_mode(self, ns, auth_mode):
        # type: (ApiNamespace, str) -> typing.Text

        # supply supported auth modes to the SDK generator using the new syntax
        backend = PythonClientBackend(
            target_folder_path='output',
            args=['-w', auth_mode, '-m', 'files', '-c', 'DropboxBase', '-t', 'dropbox'])
        backend._generate_route_methods({ns})
        return backend.output_buffer_to_string()

    def _generate_module(self, ns):
        # type: (ApiNamespace) -> typing.Text

        # Exercises the full generate() path (which writes to disk) and
        # returns the contents of the generated module.
        api = Api(version='0.1b1')
        api.namespaces[ns.name] = ns
        tmp_dir = tempfile.mkdtemp()
        try:
            backend = PythonClientBackend(
                target_folder_path=tmp_dir,
                args=['-m', 'base', '-c', 'DropboxBase', '-t', 'dropbox'])
            backend.generate(api)
            with open(os.path.join(tmp_dir, 'base.py'), encoding='utf-8') as f:
                return f.read()
        finally:
            shutil.rmtree(tmp_dir)

    def test_abstract_request_signature(self):
        # type: () -> None

        # The generated abstract request() must match the concrete
        # _DropboxTransport.request signature in dropbox-sdk-python so the
        # two base classes are Liskov-compatible under mypy (see
        # dropbox/dropbox-sdk-python#511).
        route = ApiRoute('get_metadata', 1, None)
        route.set_attributes(None, None, Void(), Void(), Void(), {})
        ns = ApiNamespace('files')
        ns.add_route(route)

        result = self._generate_module(ns)

        self.assertIn(
            '@abstractmethod\n'
            '    def request(self, route, namespace, request_arg, '
            'request_binary, timeout=None):',
            result)

    def test_route_with_version_number(self):
        # type: () -> None

        route1 = ApiRoute('get_metadata', 1, None)
        route1.set_attributes(None, ':route:`get_metadata:2`', Void(), Void(), Void(), {})
        route2 = ApiRoute('get_metadata', 2, None)
        route2.set_attributes(None, None, Void(), Int32(), Void(), {})
        ns = ApiNamespace('files')
        ns.add_route(route1)
        ns.add_route(route2)

        result = self._evaluate_namespace(ns)

        expected = textwrap.dedent('''\
            def files_get_metadata(self):
                """
                :meth:`files_get_metadata_v2`

                :rtype: None
                """
                arg = None
                r = self.request(
                    files.get_metadata,
                    'files',
                    arg,
                    None,
                )
                return None

            def files_get_metadata_v2(self):
                arg = None
                r = self.request(
                    files.get_metadata_v2,
                    'files',
                    arg,
                    None,
                )
                return r

        ''')

        self.assertEqual(result, expected)

    def test_route_with_auth_mode1(self):
        # type: () -> None

        route1 = ApiRoute('get_metadata', 1, None)
        route1.set_attributes(None, ':route:`get_metadata:2`', Void(), Void(), Void(),
                              {'auth': 'app'})
        route2 = ApiRoute('get_metadata', 2, None)
        route2.set_attributes(None, None, Void(), Int32(), Void(),
                              {'auth': 'user, app'})
        ns = ApiNamespace('files')
        ns.add_route(route1)
        ns.add_route(route2)

        result = self._evaluate_namespace_with_auth_mode(ns, 'user')

        expected = textwrap.dedent('''\
            # ------------------------------------------
            # Routes in files namespace

            def files_get_metadata_v2(self):
                arg = None
                r = self.request(
                    files.get_metadata_v2,
                    'files',
                    arg,
                    None,
                )
                return r

        ''')

        self.assertEqual(result, expected)

    def test_route_with_auth_mode2(self):
        # type: () -> None

        route1 = ApiRoute('get_metadata', 1, None)
        route1.set_attributes(None, ':route:`get_metadata:2`', Void(), Void(), Void(),
                              {'auth': 'user'})
        route2 = ApiRoute('get_metadata', 2, None)
        route2.set_attributes(None, None, Void(), Int32(), Void(),
                              {'auth': 'user, app'})
        ns = ApiNamespace('files')
        ns.add_route(route1)
        ns.add_route(route2)

        result = self._evaluate_namespace_with_auth_mode(ns, 'user')

        expected = textwrap.dedent('''\
            # ------------------------------------------
            # Routes in files namespace

            def files_get_metadata(self):
                """
                :meth:`files_get_metadata_v2`

                :rtype: None
                """
                arg = None
                r = self.request(
                    files.get_metadata,
                    'files',
                    arg,
                    None,
                )
                return None

            def files_get_metadata_v2(self):
                arg = None
                r = self.request(
                    files.get_metadata_v2,
                    'files',
                    arg,
                    None,
                )
                return r

        ''')

        self.assertEqual(result, expected)

    def test_route_with_auth_mode3(self):
        # type: () -> None

        route1 = ApiRoute('get_metadata', 1, None)
        route1.set_attributes(None, ':route:`get_metadata:2`', Void(), Void(), Void(),
                             {'auth': 'app'})
        route2 = ApiRoute('get_metadata', 2, None)
        route2.set_attributes(None, None, Void(), Int32(), Void(),
                             {'auth': 'app, team'})
        ns = ApiNamespace('files')
        ns.add_route(route1)
        ns.add_route(route2)

        result = self._evaluate_namespace_with_auth_mode(ns, 'user')

        expected = textwrap.dedent('''\
            # ------------------------------------------
            # Routes in files namespace

        ''')

        self.assertEqual(result, expected)

    def test_route_with_version_number_name_conflict(self):
        # type: () -> None

        route1 = ApiRoute('get_metadata', 2, None)
        route1.set_attributes(None, None, Void(), Int32(), Void(), {})
        route2 = ApiRoute('get_metadata_v2', 1, None)
        route2.set_attributes(None, None, Void(), Void(), Void(), {})
        ns = ApiNamespace('files')
        ns.add_route(route1)
        ns.add_route(route2)

        with self.assertRaises(RuntimeError) as cm:
            self._evaluate_namespace(ns)
        self.assertEqual(
            'There is a name conflict between {!r} and {!r}'.format(route1, route2),
            str(cm.exception))

    def test_route_argument_doc_string(self):
        backend = PythonClientBackend(
            target_folder_path='output',
            args=['-m', 'files', '-c', 'DropboxBase', '-t', 'dropbox'])
        ns = ApiNamespace('files')
        self.assertEqual(backend._format_type_in_doc(ns, Int32()), 'int')
        self.assertEqual(backend._format_type_in_doc(ns, Void()), 'None')
        self.assertEqual(backend._format_type_in_doc(ns, List(String())), 'List[str]')
        self.assertEqual(backend._format_type_in_doc(ns, Nullable(String())),
                         'Nullable[str]')
        self.assertEqual(backend._format_type_in_doc(ns, Map(String(), Int32())),
                         'Map[str, int]')

    def test_route_argument_doc_uses_foreign_type_namespace(self):
        team_common = ApiNamespace('team_common')
        time_range = Struct('TimeRange', team_common, None)
        time_range.set_attributes(None, [])

        team_log = ApiNamespace('team_log')
        args = Struct('GetTeamEventsArg', team_log, None)
        args.set_attributes(None, [
            StructField(
                'time',
                Nullable(time_range),
                'Filter by time range.',
                None,
            ),
            StructField(
                'revoke_linked_app',
                List(time_range),
                None,
                None,
            ),
        ])
        route = ApiRoute('get_events', 1, None)
        route.set_attributes(None, None, args, Void(), Void(), {})
        team_log.add_route(route)

        result = self._evaluate_namespace(team_log)

        self.assertIn(
            ':param time: Filter by time range.\n'
            '    :type time: Nullable[:class:`dropbox.team_common.TimeRange`]',
            result,
        )
        self.assertNotIn(':param Nullable[:class:', result)
        self.assertIn(
            ':type revoke_linked_app: '
            'List[:class:`dropbox.team_common.TimeRange`]',
            result,
        )

        async_ns = ApiNamespace('async')
        poll_error = Struct('PollError', async_ns, None)
        poll_error.set_attributes(None, [])
        backend = PythonClientBackend(
            target_folder_path='output',
            args=['-m', 'files', '-c', 'DropboxBase', '-t', 'dropbox'],
        )
        self.assertEqual(
            backend._format_type_in_doc(team_log, poll_error),
            ':class:`dropbox.async_.PollError`',
        )

    def test_route_with_attributes_in_docstring(self):
        # type: () -> None

        route = ApiRoute('get_metadata', 1, None)
        route.set_attributes(None, None, Void(), Void(), Void(), {
            'scope': 'events.read', 'another_attribute': 'foo'
        })
        ns = ApiNamespace('files')
        ns.add_route(route)

        result = self._evaluate_namespace(ns)
        expected = textwrap.dedent('''\
            def files_get_metadata(self):
                """
                Route attributes:
                    scope: events.read
                    another_attribute: foo

                :rtype: None
                """
                arg = None
                r = self.request(
                    files.get_metadata,
                    'files',
                    arg,
                    None,
                )
                return None

        ''')
        self.assertEqual(result, expected)

    def test_route_with_none_attribute_in_docstring(self):
        # type: () -> None

        route = ApiRoute('get_metadata', 1, None)
        route.set_attributes(None, None, Void(), Void(), Void(), {
            'scope': 'events.read', 'another_attribute': None
        })
        ns = ApiNamespace('files')
        ns.add_route(route)

        result = self._evaluate_namespace(ns)
        expected = textwrap.dedent('''\
            def files_get_metadata(self):
                """
                Route attributes:
                    scope: events.read

                :rtype: None
                """
                arg = None
                r = self.request(
                    files.get_metadata,
                    'files',
                    arg,
                    None,
                )
                return None

        ''')
        self.assertEqual(result, expected)

    def test_route_with_attributes_and_doc_in_docstring(self):
        # type: () -> None
        """
        In particular make sure there's spacing b/w overview and attrs.
        """

        route = ApiRoute('get_metadata', 1, None)
        route.set_attributes(None, "Test string.", Void(), Void(), Void(),
                             {'scope': 'events.read'})
        ns = ApiNamespace('files')
        ns.add_route(route)

        result = self._evaluate_namespace(ns)
        expected = textwrap.dedent('''\
            def files_get_metadata(self):
                """
                Test string.

                Route attributes:
                    scope: events.read

                :rtype: None
                """
                arg = None
                r = self.request(
                    files.get_metadata,
                    'files',
                    arg,
                    None,
                )
                return None

        ''')
        self.assertEqual(result, expected)

    def test_route_with_doc_and_attribute_and_data_types(self):
        # type: () -> None
        ns = ApiNamespace('files')
        struct = Struct('MyStruct', ns, None)
        struct.set_attributes(None, [
            StructField('field1', Int32(), None, None),
            StructField('field2', Int32(), None, None),
        ])

        route = ApiRoute('test/route', 1, None)
        route.set_attributes(
            None, "Test string.", struct, Int32(), Void(), {'scope': 'events.read'}
        )
        ns.add_route(route)

        result = self._evaluate_namespace(ns)
        expected = textwrap.dedent('''\
            def files_test_route(self,
                                 field1,
                                 field2):
                """
                Test string.

                Route attributes:
                    scope: events.read

                :type field1: int
                :type field2: int
                :rtype: int
                """
                arg = files.MyStruct(field1,
                                     field2)
                r = self.request(
                    files.test_route,
                    'files',
                    arg,
                    None,
                )
                return r

        ''')
        self.assertEqual(result, expected)

    # TODO: add more unit tests for client code generation
